/*************************************************************************
 * elec_tools_helpers.h - Electrostatic Tools helper functions.          *
 *                                                                       *
 * Copyright (C) 2014-2015 by Laboratory of Medicinal Chemistry of       *
 * Department of Chemistry, Lomonosov Moscow State University            *
 *                                                                       *
 * Programmed by Oleg. I. Titov                                          *
 *                                                                       *
 * This file is part of the Electrostatic Tools project.                 *
 * For more information, see <http://qsar.chem.msu.ru/elec_tools/>       *
 *                                                                       *
 * Based on work:                                                        *
 * O.I. Titov, D.A. Shulga, V.A. Palyulin, N.S. Zefirov                  *
 * Description of halogen bonding on the basis of multicenter multipole  *
 * expansion // Doklady Chemistry, 2013, vol. 450, p. 139                *
 *                                                                       *
 * This program is free software; you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation version 2 of the License.                *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 ************************************************************************/

#ifndef ELEC_TOOLS_HELPERS_HEADER
#define ELEC_TOOLS_HELPERS_HEADER

#include <string>
#include <exception>

#include <boost/program_options.hpp>

namespace po = boost::program_options;

#include "comment_skipper.h"
#include "config.h"
#include "elec_tools_errors.h"

inline void add_generic_options(po::options_description& generic_opts)
{
  generic_opts.add_options()
                  ("version", "print version string")
                  ("help", "show this message");
}

inline void create_mep_fitter_options(po::options_description& all_opts, po::positional_options_description& positional, std::string out_suffix)
{
  po::options_description input_opts("Input control");
  input_opts.add_options()
                ("input,I", po::value<std::string>()->required(), "<required> input molecule file")
                ("filetype,t", po::value<std::string>()->default_value("GUESS"), "input molecule file type (extension or GUESS)")
                ("grid,G", po::value<std::string>()->required(), "<required> input reference ESP grid file")
                ("placement-rules,p", po::value<std::string>(), "multipole placement rules")
                ("orient-rules,o", po::value<std::string>(), "multipole orient rules");

  po::options_description output_opts("Output control");
  output_opts.add_options()
                 ("output,O", po::value<std::string>(), ("output molecule file. Default: <input>." + out_suffix).c_str())
                 ("log,L", po::value<std::string>(), "log output file. Default: <output>.log")
                 ("force-output,f", po::bool_switch()->default_value(false), "owerwrite existing files");

  po::options_description fit_opts("MEP fit control");
  fit_opts.add_options()
              ("algorithm,a", po::value<std::string>()->default_value("SVD"), "optimization algorithm. "
                  "Valid values are: SVD, LLT, LDLT, PartialPivLU, FullPivLU, HouseholderQR, ColPivHouseholderQR, FullPivHouseholderQR")
              ("cutoff,c", po::value<double>()->default_value(1e7, "10^7"), "condition value cutoff (only for SVD optimizer)")
              ("recalculate-topology,r", po::bool_switch()->default_value(false), "ignore specified topology information")
              ("break-equivalency,b", po::bool_switch()->default_value(false), "remove topological equivalency constraint");

  all_opts.add(input_opts);
  all_opts.add(output_opts);
  all_opts.add(fit_opts);

  positional.add("input", 1);
  positional.add("grid", 1);
  positional.add("output", 1);
}

inline bool process_generic_opts(po::variables_map& vm, po::options_description& all_opts, std::string progname)
{
  bool processed = false;
  if (vm.count("help"))
  {
    std::cout << all_opts << std::endl;
    processed = true;
  }
  if (vm.count("version"))
  {
    std::cout << progname <<" v " << project_version << std::endl;
    processed = true;
  }
  return processed;
}

inline std::string add_filename_suffix(std::string name, std::string suffix, std::string filetype)
{
  std::transform(suffix.begin(), suffix.end(), suffix.begin(), tolower);
  std::transform(filetype.begin(), filetype.end(), filetype.begin(), tolower);
  int lastindex = name.find_last_of(".");
  std::string result = name.substr(0, lastindex);
  if (!suffix.empty()) result += "." + suffix;
  if (!filetype.empty()) result += "." + filetype;
  return result;
}

inline bool check_output_exists(const std::string& filename)
{
  std::ifstream om_test(filename.c_str());
  if (om_test.good())
  {
    om_test.close();
    return true;
  } else return false;
}

// some template magic

template <class T>
struct RuleHelper
{
  std::string default_filename() const;
  std::string param_name() const;
  std::string human_name() const;
};

template <>
struct RuleHelper<MultipoleOrientRules>
{
  std::string default_filename() const { return "orient.rules"; }
  std::string param_name() const { return "orient-rules"; }
  std::string human_name() const { return "multipole orient rules"; }
};

template <>
struct RuleHelper<MultipolePlacementRules>
{
  std::string default_filename() const { return "placement.rules"; }
  std::string param_name() const { return "placement-rules"; }
  std::string human_name() const { return "multipole placement rules"; }
};

template <class RulesT>
inline
int read_rules(RulesT& rules, po::variables_map& vm, std::ostream* logstrm)
{
  RuleHelper<RulesT> helper;
  std::string paramname = helper.param_name();
  std::string filename;
  if (vm.count(paramname))
    filename = vm[paramname].as< std::string >();
  else filename = config_path + helper.default_filename();
  std::istream* fstrm;
  std::ifstream *test_strm = new std::ifstream(filename.c_str());
  if (!test_strm->good())
  {
    *logstrm << "Cannot read " << helper.human_name() << ": \"" << filename << "\". Continuing with defaults." << std::endl;
    delete test_strm;
    if (! vm.count(paramname))
    {
      std::cerr << "Default " << helper.human_name() << " file not found. Aborting." << std::endl;
      return ELEC_TOOLS_BAD_INSTALLATION;
    }
    fstrm = new std::ifstream((config_path+helper.default_filename()).c_str());
  } else fstrm = test_strm;

  CommentSkipper skipper;
  std::istringstream strm(skipper.skip(*fstrm));

  std::string result = rules.readMe(strm);
  if( ! result.empty() )
  {
    std::cerr << "Reading " << helper.human_name() << " failed. Found unexpected: \"" << result << "\"." << std::endl;
    return ELEC_TOOLS_RULE_READ_FAILURE;
  }
  delete fstrm;
  return 0;
}

inline int open_log(std::ostream** strm, const std::string& logname)
{
  *strm = new std::ofstream(logname.c_str());
  if (!(*strm)->good())
  {
    std::cout << "Error opening log file: \"" << logname << "\". Continuing with cout." << std::endl;
    *strm = &std::cout;
  }
  return 0;
}

inline int open_output(std::ostream** strm, const std::string& filename)
{
  *strm = new std::ofstream(filename.c_str());
  if (!(*strm)->good()) {
    std::cerr << "Error opening output file: \"" << filename << "\". Aborting." << std::endl;
    return ELEC_TOOLS_UNWRITABLE_OUTPUT;
  }
  return 0;
}

inline int read_grid(Grid& grid, const std::string& filename)
{
  std::ifstream igstream(filename.c_str());
  if (!igstream.good()) {
    std::cerr << "Error opening grid file: \"" << filename << "\". Aborting." << std::endl;
    return ELEC_TOOLS_NO_GRID_FILE;
  }
  grid.fromESP(igstream);
  igstream.close();
  return 0;
}

inline int read_ob_mol(po::variables_map& vm, OpenBabel::OBMol& mol)
{
  std::ifstream molfilestrm(vm["input"].as<std::string>().c_str());
  if (!molfilestrm.good()) {
    std::cerr << "Error opening molecule file: \"" << vm["input"].as<std::string>() << "\". Aborting." << std::endl;
    return ELEC_TOOLS_NO_INPUT_FILE;
  } else
  {
    OpenBabel::OBConversion conv(&molfilestrm);
    if (vm["filetype"].as<std::string>() != "GUESS") {
      if(!conv.SetInFormat(vm["filetype"].as<std::string>().c_str()))
      {
        std::cerr << "Unsuppoted input molecule file format: \"" << vm["filetype"].as<std::string>() << "\". Aborting." << std::endl;
        return ELEC_TOOLS_BAD_MOL_FORMAT;
      }
    }
    else {
      if (!conv.SetInFormat(OpenBabel::OBConversion::FormatFromExt(vm["input"].as<std::string>().c_str()))){
        std::cerr << "Input molecule format guess failed. Aborting." << std::endl;
        return ELEC_TOOLS_BAD_MOL_FORMAT;
      }
    }

    if(!conv.Read(&mol)){
      std::cerr << "Failure reading input molecule. Aborting." << std::endl;
      return ELEC_TOOLS_MOL_READ_FAILURE;
    }
  }
  molfilestrm.close();
  return 0;
}

std::vector<int> get_unique_SMARTS_matches(const std::vector< std::string >& patterns, OpenBabel::OBMol& mol)
{
  std::vector< int > matches;
  for (int i = 0; i < patterns.size(); ++i)
  {
    OpenBabel::OBSmartsPattern patt;
    patt.Init(patterns[i]);
    patt.Match(mol);
    std::vector< std::vector< int > > hits = patt.GetUMapList();
    for (int j = 0; j < hits.size(); ++j) matches.push_back(hits[j][0]);
  }

  std::sort(matches.begin(), matches.end());
  matches.erase(std::unique(matches.begin(), matches.end()), matches.end());
  return matches;
}

#endif // ELEC_TOOLS_HELPERS_HEADER